/**********************************************************************
 **                                                                  **
 **                        cfgedit_cbrclient.cc                      **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains functions for the CfgEditCBRClient class,    **
 **  whose job is to act as an interface between CfgEdit and         **
 **  CBRServer.                                                      **
 **                                                                  **
 **  Copyright 2003 - 2006 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: cfgedit_cbrclient.cc,v 1.1.1.1 2006/07/12 13:37:54 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/un.h>
#include <pthread.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "cfgedit_cbrclient.h"
#include "cbrplanner_protocol.h"
#include "mission_expert.h"
#include "mission_expert_types.h"
#include "globals.h"
#include "EventLogging.h"
#include "string_utils.h"
#include "cnp_types.h"

using std::string;
using std::vector;
using std::ifstream;
using std::ios;

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
// Feature file related constants.
const string CfgEditCBRClient::STRING_TASK_ = "Task";
const string CfgEditCBRClient::STRING_START_X_ = "StartXCoord";
const string CfgEditCBRClient::STRING_START_Y_ = "StartYCoord";
const string CfgEditCBRClient::STRING_ROBOT_CONSTRAINTS_ = "RobotConstraints";

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constructor for CfgEditCBRClient class.
//-----------------------------------------------------------------------
CfgEditCBRClient::CfgEditCBRClient(const symbol_table<rc_chain> &rcTable) :
    CBRClient(rcTable)
{
    clientInfo_.type = CBRPLANNER_CLIENT_CFGEDIT;
    clientInfo_.pid = getpid();
    clientInfo_.name = "CfgEdit";
}

//-----------------------------------------------------------------------
// Distructor for MlabCBRClient class.
//-----------------------------------------------------------------------
CfgEditCBRClient::~CfgEditCBRClient(void)
{
}

//-----------------------------------------------------------------------
// This function uploads the FSA (tasks and triggers) names to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::saveFSANames(config_fsa_lists_t fsaLists)
{
    pthread_mutex_lock(&socketMutex_);
    sendFSANamesToServer_(fsaLists);
    pthread_mutex_unlock(&socketMutex_);
}

//-----------------------------------------------------------------------
// This function saves the new mission plan into the CBRServer.
//-----------------------------------------------------------------------
vector<int> CfgEditCBRClient::saveMissionPlan(
    vector<MExpFeature_t> globalFeatures_,
    MExpMissionTask_t task,
    double taskWeight,
    string solutionFileName)
{
    int i, j, k;
    char buf[1024];
    MExpFeature_t globalFeature;
    MExpFeature_t localFeature;
    CBRPlanner_MissionPlan_t missionplan;
    CBRPlanner_Feature_t feature;
    vector<CBRPlanner_Feature_t> features;
    vector<int> dataIndexList;

    // Describe the mission plan type.
    missionplan.type = CBRPLANNER_MISSIONPLAN_NEW;

    // Describe the mission plan rating.
    missionplan.rating = 0;

    // Open the solution file to read.
    ifstream istr(solutionFileName.c_str());

    if (istr.bad()) 
    {
        fprintf
            (stderr,
             "Error(cfgedit): CfgEditCBRClient::saveMissionPlan(). file %s not readable",
             solutionFileName.c_str());
        return dataIndexList;
    }

    istr.seekg (0, ios::end);
    missionplan.dataSize = istr.tellg();
    istr.seekg (0, ios::beg);

    // To append NULL at the end.
    missionplan.dataSize++; 

    // Read in the solution.
    missionplan.data = new char[missionplan.dataSize];
    istr.read (missionplan.data, missionplan.dataSize-1);

    // Append a null, to make it a char string.
    missionplan.data[missionplan.dataSize-1] = '\0';

    // Close the solution file.
    istr.close();

    // Number of features.
    i = 0;

    // Save the global features.
    for (j = 0; j<(int)globalFeatures_.size(); j++)
    {
        globalFeature = globalFeatures_[j];
        feature.name = globalFeature.name;
        feature.nonIndex = globalFeature.nonIndex;
        feature.weight = globalFeature.weight;

        switch (globalFeature.optionType) {
	    
        case MEXP_FEATURE_OPTION_TOGGLE:
            for (k = 0; k < ((int)(globalFeature.options.size())); k++)
            {
                if (globalFeature.selectedOption == k)
                {
                    feature.value = globalFeature.options[k].value;
                    features.push_back(feature);
                    break;
                }
            }
            break;

        case MEXP_FEATURE_OPTION_SLIDER1:
            sprintf(buf, "%.2f", globalFeature.selectedOption);
            feature.value = buf;
            features.push_back(feature);
            break;

        case MEXP_FEATURE_OPTION_SLIDER100:
        case MEXP_FEATURE_OPTION_SLIDER10:
            sprintf(buf, "%d", (int)(globalFeature.selectedOption));
            feature.value = buf;
            features.push_back(feature);
            break;
        }
    }

    // Save the task name.
    feature.name = STRING_TASK_;
    feature.value = task.name;
    feature.nonIndex = false;
    feature.weight = taskWeight;
    features.push_back(feature);

    // Save other local features.
    for (j = 0; j < ((int)(task.localFeatures.size())); j++)
    {
        localFeature = task.localFeatures[j];
        feature.name = localFeature.name;
        feature.nonIndex = localFeature.nonIndex;
        feature.weight = localFeature.weight;

        switch (localFeature.optionType) {
	    
        case MEXP_FEATURE_OPTION_TOGGLE:
            for (k = 0; k < ((int)(localFeature.options.size())); k++)
            {
                if (localFeature.selectedOption == k)
                {
                    feature.value = localFeature.options[k].value;
                    features.push_back(feature);
                    break;
                }
            }
            break;

        case MEXP_FEATURE_OPTION_SLIDER1:
            sprintf(buf, "%.2f", localFeature.selectedOption);
            feature.value = buf;
            features.push_back(feature);
            break;

        case MEXP_FEATURE_OPTION_SLIDER100:
        case MEXP_FEATURE_OPTION_SLIDER10:
            sprintf(buf, "%d", (int)(localFeature.selectedOption));
            feature.value = buf;
            features.push_back(feature);
            break;
        }
    }

    pthread_mutex_lock(&socketMutex_);
    dataIndexList = sendMissionPlanToServer_(missionplan, features);
    pthread_mutex_unlock(&socketMutex_);

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function retrieve a mission planfrom the CBRServer based on the
// features defined in the specified feature file.
//-----------------------------------------------------------------------
vector<MExpRetrievedMission_t> CfgEditCBRClient::retrieveMissionPlan(
    string featureFileName,
    vector<MExpMissionTask_t> defaultTaskList,
    vector<MExpFeature_t> defaultGlobalFeatures,
    vector<MExpFeature_t> defaultLocalFeatures,
    double taskWeight,
    int maxRating,
    string *overlayFileName,
    int *cnpMode,
    vector<MExpRobotIDs_t> *robotIDList,
    MExpCNPStatus_t *cnpStatus,
    bool *reload)
{
    typedef enum ConstraintType_t {
        CONSTRAINT_UNKNOWN,
        CONSTRAINT_TASK,
        CONSTRAINT_ROBOT
    };

    MExpRetrievedMission_t retrievedMission;
    CBRPlanner_Feature_t feature, globalFeature;
    CBRPlanner_Constraint_t constraint;
    CBRPlanner_Constraints_t genericConstraints;
    CBRPlanner_CNPStatus_t retrievedCNPStatus;
    vector<CBRPlanner_MissionPlan_t> missionplans;
    vector<CBRPlanner_Feature_t> features;
    vector<MExpRetrievedMission_t> retrievedMissions;
    vector<CBRPlanner_TaskConstraints_t> taskConstraintsList;
    vector<CBRPlanner_RobotConstraints_t> robotConstraintsList;
    vector<CBRPlanner_RobotIDs_t> retrievedRobotIDList;
    string bufString1, bufString2;
    FILE *featureFile;
    const MExpRetrievedMission_t EMPTY_RETRIEVED_MISSION;
    const vector<MExpRetrievedMission_t> EMPTY_RETRIEVED_MISSIONS;
    float selectedOption;
    char buf[1024], buf1[1024], buf2[1024], buf3[1024];
    int i, j;
    int taskId;
    int numOptions, optionType, optionNum;
    int numConstraints, constraintType = CONSTRAINT_UNKNOWN;
    int listSize, numIDs;
    int localFeatureId, globalFeatureId;
    int cnpModeInFile;
    bool dataExists = false, hadError = false;
    
    featureFile = fopen(featureFileName.c_str(), "r");
    if (featureFile == NULL)
    {
        return EMPTY_RETRIEVED_MISSIONS;
    }

    while(!feof(featureFile))
    {
        if (fscanf(featureFile, "%s", buf1) == EOF)
        {
            break;
        }

        bufString1 = buf1;

        if (bufString1 ==  MEXP_STRING_OVERLAY)
        {
            // Save the overlay name specified.
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            *overlayFileName = buf2;
        }
        else if (bufString1 ==  MEXP_STRING_CNP)
        {
            // Check the mode of CNP
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            bufString2 = buf2;

            if (isdigit(bufString2[0]))
            {
                cnpModeInFile = atoi(buf2);
            }
            else
            {
                for (i = 0; i < NUM_CNP_MODES; i++)
                {
                    if (bufString2 == CNP_MODE_STRING[i])
                    {
                        cnpModeInFile = i;
                        break;
                    }
                }
            }

            if (cnpModeInFile != *cnpMode)
            {
                *cnpMode = cnpModeInFile;
                *reload = true;
                fclose(featureFile);
                return EMPTY_RETRIEVED_MISSIONS;
            }

            *reload = false;

        }
        else if (bufString1 == MEXP_STRING_GLOBALFEATURE)
        {
            // Build a list of the global preferences.  
            // The task entries will append them to each feature list.

            // Get the preference ID (e.g., NumberOfRobots, Time, etc.)
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            globalFeatureId = atoi(buf2);

            // Get the selected option.
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            selectedOption = atof(buf2);

            for (i = 0; i < ((int)(defaultGlobalFeatures.size())); i++)
            {
                if (defaultGlobalFeatures[i].id == globalFeatureId)
                {
                    globalFeature.name = defaultGlobalFeatures[i].name;
                    globalFeature.nonIndex = defaultGlobalFeatures[i].nonIndex;
                    globalFeature.weight = defaultGlobalFeatures[i].weight;

                    switch (defaultGlobalFeatures[i].optionType) {

                    case MEXP_FEATURE_OPTION_TOGGLE:
                        globalFeature.value = defaultGlobalFeatures[i].options[(int)selectedOption].value;
                        globalFeature.formatType = CBRPLANNER_FORMAT_STRING;
                        break;
			
                    case MEXP_FEATURE_OPTION_SLIDER1:
                        sprintf(buf, "%.2f", selectedOption);
                        globalFeature.value = buf;
                        globalFeature.formatType = CBRPLANNER_FORMAT_DOUBLE;
                        break;

                    case MEXP_FEATURE_OPTION_SLIDER100:
                    case MEXP_FEATURE_OPTION_SLIDER10:
                        sprintf(buf, "%d", (int)selectedOption);
                        globalFeature.value = buf;
                        globalFeature.formatType = CBRPLANNER_FORMAT_INT;
                        break;
                    }

                    break;
                }
            }

            features.push_back(globalFeature);
        }
        else if (bufString1 == MEXP_STRING_TASK)
        {
            feature.name = STRING_TASK_;
            feature.nonIndex = false;
            feature.weight = taskWeight;
            feature.formatType = CBRPLANNER_FORMAT_STRING;

            // Get the task ID (e.g., Biohazrd, EOD, etc.)
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            taskId = atoi(buf2);

            for (i = 0; i < ((int)(defaultTaskList.size())); i++)
            {
                if (defaultTaskList[i].id == taskId)
                {
                    feature.value = defaultTaskList[i].name;
                    break;
                }
            }
	    
            features.push_back(feature);

            dataExists = true;
        }
        else if (bufString1 == MEXP_STRING_COORDINATE)
        {
            feature.nonIndex = true;
            feature.weight = 0.0;
            feature.formatType = CBRPLANNER_FORMAT_DOUBLE;

            // Get the coordinate.
            if (fscanf(featureFile, "%s %s", buf2, buf3) == EOF)
            {
                hadError = true;
                break;
            }

            feature.name = STRING_START_X_;
            feature.value = buf2;
            features.push_back(feature);

            feature.name = STRING_START_Y_;
            feature.value = buf3;
            features.push_back(feature);

            dataExists = true;
        }
        else if (bufString1 == MEXP_STRING_FEATURE)
        {
            // Get the feature ID.
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            localFeatureId = atoi(buf2);

            // Get the feature option being selected.
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            selectedOption = atof(buf2);

            for (i = 0; i < ((int)(defaultLocalFeatures.size())); i++)
            {
                if (defaultLocalFeatures[i].id == localFeatureId)
                {
                    feature.name = defaultLocalFeatures[i].name;
                    feature.nonIndex = defaultLocalFeatures[i].nonIndex;
                    feature.weight = defaultLocalFeatures[i].weight;

                    switch (defaultLocalFeatures[i].optionType) {
			
                    case MEXP_FEATURE_OPTION_TOGGLE:
                        feature.value = defaultLocalFeatures[i].options[(int)selectedOption].value;
                        feature.formatType = CBRPLANNER_FORMAT_STRING;
                        break;

                    case MEXP_FEATURE_OPTION_SLIDER1:
                        sprintf(buf, "%.2f", selectedOption);
                        feature.value = buf;
                        feature.formatType = CBRPLANNER_FORMAT_DOUBLE;
                        break;

                    case MEXP_FEATURE_OPTION_SLIDER100:
                    case MEXP_FEATURE_OPTION_SLIDER10:
                        sprintf(buf, "%d", (int)selectedOption);
                        feature.value = buf;
                        feature.formatType = CBRPLANNER_FORMAT_INT;
                        break;
                    }

                    features.push_back(feature);
                    break;
                }
            }

            dataExists = true;
        }
        else if ((bufString1 == MEXP_STRING_TASK_CONSTRAINTS) ||
                 (bufString1 == MEXP_STRING_ROBOT_CONSTRAINTS))
        {
            if (bufString1 == MEXP_STRING_TASK_CONSTRAINTS)
            {
                constraintType = CONSTRAINT_TASK;
            }
            else if (bufString1 == MEXP_STRING_ROBOT_CONSTRAINTS)
            {
                constraintType = CONSTRAINT_ROBOT;
            }

            genericConstraints.constraints.clear();

            // Get the task/robot ID
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            genericConstraints.id = atoi(buf2);

            // Get the number of constraints
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            numConstraints = atoi(buf2);

            // Get the name of the task/robot
            if (fscanf(featureFile, "%s", buf2) == EOF)
            {
                hadError = true;
                break;
            }
            bufString2 = buf2;
            // Remove '"'
            bufString2 = removeCharInString(bufString2, '\"');
            genericConstraints.name = bufString2;

            for (i = 0; i < numConstraints; i++)
            {
                // Make sure it is the constraint first.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                bufString2 = buf2;

                if ((bufString2 != MEXP_STRING_TASK_CONSTRAINT) &&
                    (bufString2 != MEXP_STRING_ROBOT_CONSTRAINT))
                {
                    hadError = true;
                    break;
                }

                // Get the constraint ID.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                constraint.id = atoi(buf2);

                // Get the option being selected.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                selectedOption = atof(buf2);

                // Get the number of options.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                numOptions = atoi(buf2);

                // Get the option types.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                optionType = atoi(buf2);

                // Get the constraint name.
                if (fscanf(featureFile, "%s", buf2) == EOF)
                {
                    hadError = true;
                    break;
                }
                bufString2 = buf2;
                // Remove '"'
                bufString2 = removeCharInString(bufString2, '\"');
                constraint.name = bufString2;

                switch (optionType) {

                case MEXP_FEATURE_OPTION_SLIDER1:
                    sprintf(buf, "%.2f", selectedOption);
                    constraint.value = buf;
                    constraint.formatType = CBRPLANNER_FORMAT_DOUBLE;
                    break;

                case MEXP_FEATURE_OPTION_SLIDER100:
                case MEXP_FEATURE_OPTION_SLIDER10:
                    sprintf(buf, "%d", (int)selectedOption);
                    constraint.value = buf;
                    constraint.formatType = CBRPLANNER_FORMAT_INT;
                    break;

                case MEXP_FEATURE_OPTION_TOGGLE:
                    // Parse though the options until it finds the selected option.
                    for (j = 0; j < numOptions; j++)
                    {
                        // Get the option string
                        if (fscanf(featureFile, "%s", buf2) == EOF)
                        {
                            hadError = true;
                            break;
                        }
                        bufString2 = buf2;

                        if (bufString2 != MEXP_STRING_OPTION)
                        {
                            hadError = true;
                            break;
                        }

                        // Get the option number
                        if (fscanf(featureFile, "%s", buf2) == EOF)
                        {
                            hadError = true;
                            break;
                        }
                        optionNum = atoi(buf2);

                        // Get the option value
                        if (fscanf(featureFile, "%s", buf2) == EOF)
                        {
                            hadError = true;
                            break;
                        }

                        if (optionNum == ((int)selectedOption))
                        {
                            bufString2 = buf2;
                            bufString2 = removeCharInString(bufString2, '\"');
                            constraint.value = bufString2;
                        }
                    }

                    constraint.formatType = CBRPLANNER_FORMAT_STRING;
                    break;

                default:
                    // Unknown option type.
                    hadError = true;
                    break;
                }

                if (hadError)
                {
                    break;
                }

                genericConstraints.constraints.push_back(constraint);
            }

            switch (constraintType) {

            case CONSTRAINT_TASK:
                taskConstraintsList.push_back(
                    (CBRPlanner_TaskConstraints_t)genericConstraints);
                break;

            case CONSTRAINT_ROBOT:
                robotConstraintsList.push_back(
                    (CBRPlanner_RobotConstraints_t)genericConstraints);
                break;
            }
        }
    }

    fclose(featureFile);

    if ((!dataExists)||(hadError))
    {
        return EMPTY_RETRIEVED_MISSIONS;
    }

    // Get the mission plan data from the server.
    pthread_mutex_lock(&socketMutex_);
    gEventLogging->start("CfgEditCBRClient Retrieving mission plan from CBRServer");

    switch (*cnpMode) {

    case CNP_MODE_DISABLED:
        missionplans = retrieveMissionPlansFromServer_(
            features,
            maxRating);
        break;

    case CNP_MODE_PREMISSION:
        missionplans = retrieveMissionPlansFromServerPremissionCNP_(
            features,
            maxRating,
            taskConstraintsList,
            robotConstraintsList);
        retrievedRobotIDList = retrieveCurrentRobotIDListFromServer_();
        retrievedCNPStatus = retrieveCurrentCNPStatusFromServer_();
        listSize = retrievedRobotIDList.size();
        robotIDList->resize(listSize);

        for (i = 0; i < listSize; i++)
        {
            numIDs = retrievedRobotIDList[i].robotIDs.size();
            (*robotIDList)[i].robotIDs.resize(numIDs);

            for (j = 0; j < numIDs; j++)
            {
                (*robotIDList)[i].robotIDs[j] = retrievedRobotIDList[i].robotIDs[j];
            }
        }

        cnpStatus->status = retrievedCNPStatus.status;
        cnpStatus->errorMsg = retrievedCNPStatus.errorMsg;
        break;

    case CNP_MODE_RUNTIME:
        missionplans = retrieveMissionPlansFromServerRuntimeCNP_(
            features,
            maxRating,
            robotConstraintsList);
        retrievedRobotIDList = retrieveCurrentRobotIDListFromServer_();
        listSize = retrievedRobotIDList.size();
        robotIDList->resize(listSize);

        for (i = 0; i < listSize; i++)
        {
            numIDs = retrievedRobotIDList[i].robotIDs.size();
            (*robotIDList)[i].robotIDs.resize(numIDs);

            for (j = 0; j < numIDs; j++)
            {
                (*robotIDList)[i].robotIDs[j] = retrievedRobotIDList[i].robotIDs[j];
            }
        }
        break;

    case CNP_MODE_PREMISSION_AND_RUNTIME:
    default:
        fprintf
            (stderr,
             "Warning: CfgEditCBRClient::retrieveMissionPlan(). Unknown CNP mode.\a\n");
        missionplans = retrieveMissionPlansFromServer_(
            features,
            maxRating);
        break;
    }

    gEventLogging->end("CfgEditCBRClient Retrieving mission plan from CBRServer");
    pthread_mutex_unlock(&socketMutex_);

    for (i = 0; i < ((int)(missionplans.size())); i++)
    {
        retrievedMission = EMPTY_RETRIEVED_MISSION;
        retrievedMission.rating = missionplans[i].rating;
        retrievedMission.dataIndexList = missionplans[i].dataIndexList;
        retrievedMission.acdlSolution = missionplans[i].data;
        retrievedMissions.push_back(retrievedMission);
    }
    return retrievedMissions;
}

//-----------------------------------------------------------------------
// This function asks the CBRServer to fix the mission plan.
//-----------------------------------------------------------------------
vector<MExpRetrievedMission_t> CfgEditCBRClient::fixMissionPlan(string missionPlanFileName)
{
    CBRPlanner_MissionPlan_t missionplan;
    CBRPlanner_Feature_t feature;
    MExpRetrievedMission_t retrievedMission;
    vector<MExpRetrievedMission_t> retrievedMissions;
    bool fixed = false;

    // Describe the mission plan type.
    missionplan.type = CBRPLANNER_MISSIONPLAN_NEW;

    // Describe the mission plan rating.
    missionplan.rating = 0;

    // Open the solution file to read.
    ifstream istr(missionPlanFileName.c_str());

    if (istr.bad()) 
    {
        fprintf
            (stderr,
             "Error: CfgEditCBRClient::saveMissionPlan(). file %s not readable",
             missionPlanFileName.c_str());
        return retrievedMissions;
    }

    istr.seekg (0, ios::end);
    missionplan.dataSize = istr.tellg();
    istr.seekg (0, ios::beg);

    // To append NULL at the end.
    missionplan.dataSize++; 

    // Read in the solution.
    missionplan.data = new char[missionplan.dataSize];
    istr.read (missionplan.data, missionplan.dataSize-1);

    // Append a null, to make it a char string.
    missionplan.data[missionplan.dataSize-1] = '\0';

    // Close the solution file.
    istr.close();

    pthread_mutex_lock(&socketMutex_);

    // First, send a flag to notify the server that the mission plan needs to be fixed.
    sendCommandToServer_(CBRPCOM_CMD_MISSIONPLAN_FIX);

    // Next, send the mission plan. The mission plan data is composed of the type flag,
    // rating, data size, and the data itselt.
    writeInteger_(shSocketFd_, missionplan.type);
    writeInteger_(shSocketFd_, missionplan.rating);
    writeInteger_(shSocketFd_, missionplan.dataSize);
    write(shSocketFd_, missionplan.data, missionplan.dataSize);

    // Clear the data.
    delete [] missionplan.data;
    missionplan.data = NULL;

    // Now retrieve the mission plan from the server. Make repairment was successful.
    fixed = readBoolean_(shSocketFd_);

    if (fixed)
    {
        missionplan = retrieveMissionPlanFromServer_();
    }

    pthread_mutex_unlock(&socketMutex_);

    if (!fixed)
    {
        // It's an empty mission.
        return retrievedMissions;
    }

    retrievedMission.rating = missionplan.rating;
    retrievedMission.dataIndexList = missionplan.dataIndexList;
    retrievedMission.acdlSolution = missionplan.data;
    retrievedMissions.push_back(retrievedMission);

    return retrievedMissions;
}

//-----------------------------------------------------------------------
// This function sends a mission plan to the server.
//-----------------------------------------------------------------------
vector<int> CfgEditCBRClient::sendMissionPlanToServer_(
    CBRPlanner_MissionPlan_t missionplan,
    vector<CBRPlanner_Feature_t> features)
{
    int numFeatures, nameSize, valueSize, dataIndexListSize;
    int i, dataIndex;
    int formatType;
    vector<int> dataIndexList;

    // First, send a flag to notify the server that the mission plan will be sent.
    sendCommandToServer_(CBRPCOM_CMD_MISSIONPLAN_SAVE);

    // Next, send the mission plan. The mission plan data is composed of the type flag,
    // rating, data size, and the data itselt.
    writeInteger_(shSocketFd_, missionplan.type);
    writeInteger_(shSocketFd_, missionplan.rating);
    writeInteger_(shSocketFd_, missionplan.dataSize);
    write(shSocketFd_, missionplan.data, missionplan.dataSize);

    // Send the number of features
    numFeatures = features.size();
    writeInteger_(shSocketFd_, numFeatures);

    for (i = 0; i < numFeatures; i++)
    {
        // Send the length of the name
        nameSize = features[i].name.size();
        writeInteger_(shSocketFd_, nameSize);

        // Send the name
        write(shSocketFd_, features[i].name.c_str(), nameSize);
	
        // Send the length of the value
        valueSize = features[i].value.size();
        writeInteger_(shSocketFd_, valueSize);

        // Send the value
        write(shSocketFd_, features[i].value.c_str(), valueSize);

        // Send the nonIndex flag.
        writeBoolean_(shSocketFd_, features[i].nonIndex);

        // Send the weight.
        writeDouble_(shSocketFd_, features[i].weight);

        // Send the format type
        formatType = features[i].formatType;
        writeInteger_(shSocketFd_, formatType);
    }

    // Send zero as the number of robots
    writeInteger_(shSocketFd_, 0);

    // Get the new list of data indexes.
    dataIndexListSize = readInteger_(shSocketFd_);

    for (i = 0; i < dataIndexListSize; i++)
    {
        dataIndex = readInteger_(shSocketFd_);
        dataIndexList.push_back(dataIndex);
    }

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function retrieves mission plans afer sending features to the server.
//-----------------------------------------------------------------------
vector<CBRPlanner_MissionPlan_t> CfgEditCBRClient::retrieveMissionPlansFromServer_(
    vector<CBRPlanner_Feature_t> features,
    int maxRating)
{
    vector<CBRPlanner_MissionPlan_t> missionplans;

    // First, send a flag to notify the server that the mission plan is going
    // to be retrieved.
    sendCommandToServer_(CBRPCOM_CMD_MISSIONPLAN_RETRIEVE);

    // Send the features
    sendFeaturesToServer_(features);

    // Send the maximum suitability rating
    write(shSocketFd_, &maxRating, sizeof(int));

    // Now retrieve the mission plan from the server.
    missionplans = receiveMissionPlansFromServer_();

    return missionplans;
}

//-----------------------------------------------------------------------
// This function retrieves mission plans afer sending features to the server,
// with premission-CNP enabled.
//-----------------------------------------------------------------------
vector<CBRPlanner_MissionPlan_t> CfgEditCBRClient::retrieveMissionPlansFromServerPremissionCNP_(
    vector<CBRPlanner_Feature_t> features,
    int maxRating,
    vector<CBRPlanner_TaskConstraints_t> taskConstraintsList,
    vector<CBRPlanner_RobotConstraints_t> robotConstraintsList)
{
    vector<CBRPlanner_MissionPlan_t> missionplans;
    vector<CBRPlanner_Constraints_t> genericConstraintsList;

    // First, send a flag to notify the server that the mission plan is going
    // to be retrieved.
    sendCommandToServer_(CBRPCOM_CMD_MISSIONPLAN_RETRIEVE_CNP_PREMISSION);

    // Send the features
    sendFeaturesToServer_(features);

    // Send the maximum suitability rating
    write(shSocketFd_, &maxRating, sizeof(int));

    // Send the task constraints.
    genericConstraintsList = (vector<CBRPlanner_Constraints_t>)taskConstraintsList;
    sendConstraintsToServer_(genericConstraintsList);

    // Send the robot constraints.
    genericConstraintsList = (vector<CBRPlanner_Constraints_t>)robotConstraintsList;
    sendConstraintsToServer_(genericConstraintsList);

    // Now retrieve the mission plan from the server.
    missionplans = receiveMissionPlansFromServer_();

    return missionplans;
}

//-----------------------------------------------------------------------
// This function retrieves mission plans afer sending features to the server.
//-----------------------------------------------------------------------
vector<CBRPlanner_MissionPlan_t> CfgEditCBRClient::retrieveMissionPlansFromServerRuntimeCNP_(
    vector<CBRPlanner_Feature_t> features,
    int maxRating,
    vector<CBRPlanner_RobotConstraints_t> robotConstraintsList)
{
    vector<CBRPlanner_MissionPlan_t> missionplans;
    vector<CBRPlanner_Constraints_t> genericConstraintsList;

    // First, send a flag to notify the server that the mission plan is going
    // to be retrieved.
    sendCommandToServer_(CBRPCOM_CMD_MISSIONPLAN_RETRIEVE_CNP_RUNTIME);

    // Send the features
    sendFeaturesToServer_(features);

    // Send the maximum suitability rating
    write(shSocketFd_, &maxRating, sizeof(int));

    // Send the robot constraints.

    genericConstraintsList = (vector<CBRPlanner_Constraints_t>)robotConstraintsList;
    sendConstraintsToServer_(genericConstraintsList);

    // Now retrieve the mission plan from the server.
    missionplans = receiveMissionPlansFromServer_();

    return missionplans;
}

//-----------------------------------------------------------------------
// This function retrieves mission plans
//-----------------------------------------------------------------------
vector<CBRPlanner_MissionPlan_t> CfgEditCBRClient::receiveMissionPlansFromServer_(void)
{
    CBRPlanner_MissionPlan_t missionplan;
    vector<CBRPlanner_MissionPlan_t> missionplans;
    int i, numPlans;
    const bool DEBUG_RECEIVE_MISSION_PLANS_FROM_SERVER = false;

    // Now retrieve the mission plan from the server. Get the number of the plans.
    numPlans = readInteger_(shSocketFd_);

    for (i = 0; i < numPlans; i++)
    {
        missionplan = retrieveMissionPlanFromServer_();

        if (DEBUG_RECEIVE_MISSION_PLANS_FROM_SERVER)
        {
            fprintf(stderr, "Received Plan (%d)\n", i);
            fprintf(stderr, "%s\n\n", missionplan.data);
        }

        missionplans.push_back(missionplan);
    }

    return missionplans;
}

//-----------------------------------------------------------------------
// This function retrieves Robot ID lists.
//-----------------------------------------------------------------------
vector<CBRPlanner_RobotIDs_t> CfgEditCBRClient::retrieveCurrentRobotIDListFromServer_(void)
{
    vector<CBRPlanner_RobotIDs_t> robotIDList;
    int i, j, listSize, numIDs, id;
    const bool DEBUG = false;

    // First, send the flag
    sendCommandToServer_(CBRPCOM_CMD_ROBOT_ID_LIST_RETRIEVE);

    listSize = readInteger_(shSocketFd_);
    robotIDList.resize(listSize);

    for (i = 0; i < listSize; i++)
    {
        numIDs = readInteger_(shSocketFd_);
        robotIDList[i].robotIDs.resize(numIDs);

        for (j = 0; j < numIDs; j++)
        {
            id = readInteger_(shSocketFd_);
            robotIDList[i].robotIDs[j] = id;
        }
    }

    if (DEBUG)
    {
        fprintf(stderr, "Received Robot IDs\n");

        for (i = 0; i < (int)(robotIDList.size()); i++)
        {
            fprintf(stderr, "LIST(%d)\n", i);

            for (j = 0; j < (int)(robotIDList[i].robotIDs.size()); j++)
            {
                fprintf(stderr, "%d. Robot ID: %d\n", j, robotIDList[i].robotIDs[j]);
            }
        }
    }

    return robotIDList;
}

//-----------------------------------------------------------------------
// This function retrieves CNP status and error messages.
//-----------------------------------------------------------------------
CBRPlanner_CNPStatus_t CfgEditCBRClient::retrieveCurrentCNPStatusFromServer_(void)
{
    CBRPlanner_CNPStatus_t cnpStatus;
    string errorMsg = EMPTY_STRING_, debugMsg;
    int status, dataSize;
    char *data = NULL;
    const bool DEBUG = false;

    // First, send the flag
    sendCommandToServer_(CBRPCOM_CMD_CNP_STATUS_RETRIEVE);

    // Get the status
    status = readInteger_(shSocketFd_);

    // Get the message size.
    dataSize = readInteger_(shSocketFd_);

    // Get the message (if it has a content).
    if (dataSize > 0)
    {
        data = new char[dataSize+1];
        readNBytes_(shSocketFd_, (unsigned char *)data, dataSize);

        errorMsg = data;
        errorMsg[dataSize] = '\0';
        delete [] data;
        data = NULL;
    }

    cnpStatus.status = status;
    cnpStatus.errorMsg = errorMsg;

    if (DEBUG)
    {
        fprintf(stderr, "Retreived CNP Status: %d\n", cnpStatus.status);

        if (cnpStatus.errorMsg == EMPTY_STRING_)
        {
            debugMsg = "(empty)";
        }
        else
        {
            debugMsg = cnpStatus.errorMsg;
        }

        fprintf(stderr, "Error Message: %s\n", debugMsg.c_str());
    }

    return cnpStatus;
}

//-----------------------------------------------------------------------
// This function sends features to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendFeaturesToServer_(vector<CBRPlanner_Feature_t> features)
{
    int numFeatures, nameSize, valueSize;
    int i;
    int formatType;
    double weight;
    bool nonIndex = false;

    // Send the number of features
    numFeatures = features.size();
    writeInteger_(shSocketFd_, numFeatures);

    for (i = 0; i < numFeatures; i++)
    {
        // Send the length of the name
        nameSize = features[i].name.size();
        writeInteger_(shSocketFd_, nameSize);

        // Send the name
        write(shSocketFd_, features[i].name.c_str(), nameSize);
	
        // Send the length of the value
        valueSize = features[i].value.size();
        writeInteger_(shSocketFd_, valueSize);

        // Send the value
        write(shSocketFd_, features[i].value.c_str(), valueSize);

        // Send the nonIndex flag.
        nonIndex = features[i].nonIndex;
        writeBoolean_(shSocketFd_, nonIndex);

        // Send the weight.
        weight = features[i].weight;
        write(shSocketFd_, &weight, sizeof(double));

        // Send the format type
        formatType = features[i].formatType;
        writeInteger_(shSocketFd_, formatType);
    }
}

//-----------------------------------------------------------------------
// This function sends constraints to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendConstraintsToServer_(vector<CBRPlanner_Constraints_t> constraintsList)
{
    CBRPlanner_Constraint_t constraint;
    int listSize, numConstraints, nameSize, valueSize;
    int i, j;
    int id;
    int formatType;

    // Send the size of the list
    listSize = constraintsList.size();
    writeInteger_(shSocketFd_, listSize);

    for (i = 0; i < listSize; i++)
    {
        // Send the ID of the task/robot
        id = constraintsList[i].id;
        writeInteger_(shSocketFd_, id);

        // Send the length of the name
        nameSize = constraintsList[i].name.size();
        writeInteger_(shSocketFd_, nameSize);

        // Send the name
        write(shSocketFd_, constraintsList[i].name.c_str(), nameSize);
	
        // Send the number of constraints
        numConstraints = constraintsList[i].constraints.size();
        writeInteger_(shSocketFd_, numConstraints);

        for (j = 0; j < numConstraints; j++)
        {
            constraint = constraintsList[i].constraints[j];

            // Send the length of the name
            nameSize = constraint.name.size();
            writeInteger_(shSocketFd_, nameSize);

            // Send the name
            write(shSocketFd_, constraint.name.c_str(), nameSize);
	
            // Send the length of the value
            valueSize = constraint.value.size();
            writeInteger_(shSocketFd_, valueSize);

            // Send the value
            write(shSocketFd_, constraint.value.c_str(), valueSize);

            // Send the ID
            id = constraint.id;
            writeInteger_(shSocketFd_, id);

            // Send the format type
            formatType = constraint.formatType;
            writeInteger_(shSocketFd_, formatType);
        }
    }
}

//-----------------------------------------------------------------------
// This function retrieves a mission plan afer sending features to the server.
//-----------------------------------------------------------------------
CBRPlanner_MissionPlan_t CfgEditCBRClient::retrieveMissionPlanFromServer_(void)
{
    CBRPlanner_MissionPlan_t missionplan;
    int i, dataIndexListSize;
    int dataIndex;

    // Receive the mission type.
    missionplan.type = readInteger_(shSocketFd_);

    // Receive the rating.
    missionplan.rating = readInteger_(shSocketFd_);

    // Receive the list of the data indexes.
    missionplan.dataIndexList.clear();
    dataIndexListSize = readInteger_(shSocketFd_);

    for (i = 0; i < dataIndexListSize; i++)
    {
        dataIndex = readInteger_(shSocketFd_);
        missionplan.dataIndexList.push_back(dataIndex);
    }

    // Receive the data.
    missionplan.dataSize = readInteger_(shSocketFd_);
    missionplan.data = new char[missionplan.dataSize+1];
    readNBytes_(shSocketFd_, (unsigned char *)(missionplan.data), missionplan.dataSize);
    missionplan.data[missionplan.dataSize] = '\0';

    return missionplan;
}

//-----------------------------------------------------------------------
// This function sends the FSA (tasks and triggers) names to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendFSANamesToServer_(config_fsa_lists_t fsaLists)
{
    int i, numTasks, numTriggers, nameSize;

    // First, send a flag notify the server that fsa lists will be sent.
    sendCommandToServer_(CBRPCOM_CMD_FSANAMES);

    // Send the number of tasks.
    numTasks = fsaLists.taskList.size();
    write(shSocketFd_, &numTasks, sizeof(int));

    for (i = 0; i < numTasks; i++)
    {
        // Send the length of the task name.
        nameSize = fsaLists.taskList[i].name.size();
        write(shSocketFd_, &nameSize, sizeof(int));

        // Send the name of the task
        write(shSocketFd_, fsaLists.taskList[i].name.c_str(), nameSize);
    }

    // Send the number of triggers.
    numTriggers = fsaLists.triggerList.size();
    write(shSocketFd_, &numTriggers, sizeof(int));

    for (i = 0; i < numTriggers; i++)
    {
        // Send the length of the trigger name.
        nameSize = fsaLists.triggerList[i].name.size();
        write(shSocketFd_, &nameSize, sizeof(int));

        // Send the name of the trigger
        write(shSocketFd_, fsaLists.triggerList[i].name.c_str(), nameSize);
    }
}

//-----------------------------------------------------------------------
// This function sends a positive feedback (regarding the latest mission)
// to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::savePositiveFeedback(vector<int> dataIndexList)
{
    pthread_mutex_lock(&socketMutex_);
    sendFeedbackToServer_(CBRPCOM_CMD_POSITIVE_FEEDBACK, dataIndexList);
    pthread_mutex_unlock(&socketMutex_);
}

//-----------------------------------------------------------------------
// This function sends a negative feedback (regarding the latest mission)
// to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::saveNegativeFeedback(vector<int> dataIndexList)
{
    pthread_mutex_lock(&socketMutex_);
    sendFeedbackToServer_(CBRPCOM_CMD_NEGATIVE_FEEDBACK, dataIndexList);
    pthread_mutex_unlock(&socketMutex_);
}

//-----------------------------------------------------------------------
// This function sends a feedback (regarding the latest mission) to the
// server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendFeedbackToServer_(const unsigned char cmd, vector<int> dataIndexList)
{
    int i, dataIndexListSize;

    sendCommandToServer_(cmd);

    dataIndexListSize = dataIndexList.size();
    write(shSocketFd_, &dataIndexListSize, sizeof(int));

    for (i = 0; i < dataIndexListSize; i++)
    {
        write(shSocketFd_, &(dataIndexList[i]), sizeof(int));
    }
}

//-----------------------------------------------------------------------
// This function retrieves logfile info from the server.
//-----------------------------------------------------------------------
MExpLogfileInfo_t CfgEditCBRClient::retrieveLogfileInfoFromServer_(void)
{
    MExpLogfileInfo_t logfileInfo;
    int dataSize;
    char *data = NULL;

    dataSize = readInteger_(shSocketFd_);
    data = new char[dataSize+1];
    readNBytes_(shSocketFd_, (unsigned char *)data, dataSize);

    logfileInfo.name = data;
    logfileInfo.name[dataSize] = '\0';
    delete [] data;
    data = NULL;

    return logfileInfo;
}

//-----------------------------------------------------------------------
// This function retrieves state info from the server.
//-----------------------------------------------------------------------
vector<MExpStateInfo_t> CfgEditCBRClient::retrieveStateInfoFromServer_(void)
{
    MExpStateInfo_t stateInfo;
    vector<MExpStateInfo_t> stateInfoList;
    int i, dataSize, numStateInfo, nameSize;
    char *data, *dataStart = NULL, *nameBuf = NULL;

    dataSize = readInteger_(shSocketFd_);
    data = new char[dataSize];
    dataStart = data;
    readNBytes_(shSocketFd_, (unsigned char *)data, dataSize);

    // Copy the size of the list.
    memcpy(&numStateInfo, data, sizeof(int));
    data += sizeof(int);

    for (i = 0; i < numStateInfo; i++)
    {
        // Copy the size of the FSA name.
        memcpy(&nameSize, data, sizeof(int));
        data += sizeof(int);

        // Copy the FSA name itself.
        nameBuf = new char[nameSize+1];
        memcpy(nameBuf, data, nameSize);
        data += nameSize;
        nameBuf[nameSize] = '\0';
        stateInfo.fsaName = nameBuf;
        delete nameBuf;
        nameBuf = NULL;

        // Copy the size of the state name.
        memcpy(&nameSize, data, sizeof(int));
        data += sizeof(int);

        // Copy the state name itself.
        nameBuf = new char[nameSize+1];
        memcpy(nameBuf, data, nameSize);
        data += nameSize;
        nameBuf[nameSize] = '\0';
        stateInfo.stateName = nameBuf;
        delete nameBuf;
        nameBuf = NULL;

        stateInfoList.push_back(stateInfo);
    }

    data = dataStart;
    delete [] data;
    data = NULL;

    return stateInfoList;
}

//-----------------------------------------------------------------------
// This function sends logfile info to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendLogfileInfoToServer_(
    vector<MExpLogfileInfo_t> logfileInfoList)
{
    char *dummyData;
    int i, numLogfileInfoList, dataSize;

    // Read a dummy data first.
    dataSize = readInteger_(shSocketFd_);
    dummyData = new char[dataSize];
    readNBytes_(shSocketFd_, (unsigned char *)dummyData, dataSize);
    delete [] dummyData;

    sendCommandToServer_(CBRPCOM_CMD_LOGFILE_INFO_RETRIEVE);

    numLogfileInfoList = logfileInfoList.size();
    writeInteger_(shSocketFd_, numLogfileInfoList);

    for (i = 0; i < numLogfileInfoList; i++)
    {
        dataSize = logfileInfoList[i].name.size();
        writeInteger_(shSocketFd_, dataSize);
        write(shSocketFd_, logfileInfoList[i].name.c_str(), dataSize);
    }
}

//-----------------------------------------------------------------------
// This function sends overlay info to the server.
//-----------------------------------------------------------------------
void CfgEditCBRClient::sendOverlayInfoToServer_(string overlayFileName)
{
    char *dummyData;
    int dataSize;

    // Read a dummy data first.
    dataSize = readInteger_(shSocketFd_);
    dummyData = new char[dataSize];
    readNBytes_(shSocketFd_, (unsigned char *)dummyData, dataSize);
    delete [] dummyData;

    sendCommandToServer_(CBRPCOM_CMD_OVERLAY_INFO_RETRIEVE);

    dataSize = overlayFileName.size();
    writeInteger_(shSocketFd_, dataSize);
    write(shSocketFd_, overlayFileName.c_str(), dataSize);
}

//-----------------------------------------------------------------------
// This function read and process the server data.
//-----------------------------------------------------------------------
void CfgEditCBRClient::readAndProcessServerData_(void)
{
    MExpLogfileInfo_t logfileInfo;
    vector<MExpStateInfo_t> stateInfoList;
    vector<MExpLogfileInfo_t> logfileInfoList;
    string overlayFileName;
    unsigned char oneByteBuf[1];
    const bool DEBUG_READ_AND_PROCESS_SERVER_DATA = false;

    readNBytes_(shSocketFd_, (unsigned char *)&oneByteBuf, 1);

    if (DEBUG_READ_AND_PROCESS_SERVER_DATA)
    {
        fprintf(
            stderr,
            "CfgEditCBRClient::readAndProcessServerData_(): Byte [%x] receieved.\n",
            oneByteBuf[0]);
    }

    switch (oneByteBuf[0]) {

    case CBRPCOM_CMD_LOGFILE_INFO_SAVE:
        logfileInfo = retrieveLogfileInfoFromServer_();
        if (gMExp != NULL)
        {
            gMExp->saveLogfileInfo(logfileInfo);
        }
        break;

    case CBRPCOM_CMD_LOGFILE_INFO_RETRIEVE:
        if (gMExp != NULL)
        {
            logfileInfoList  = gMExp->getLogfileInfoList();
        }
        sendLogfileInfoToServer_(logfileInfoList);
        break;

    case CBRPCOM_CMD_OVERLAY_INFO_RETRIEVE:
        if (gMExp != NULL)
        {
            overlayFileName  = gMExp->getOverlayFileName();
        }
        sendOverlayInfoToServer_(overlayFileName);
        break;

    case CBRPCOM_CMD_REPORT_CURRENT_STATE:
        stateInfoList = retrieveStateInfoFromServer_();
        if (gMExp != NULL)
        {
            gMExp->highlightStates(stateInfoList);
        }
        break;

    default:
        break;
    }
}

//-----------------------------------------------------------------------
// This function sends the index of the highlighted state.
//-----------------------------------------------------------------------
void CfgEditCBRClient::saveHighlightedStateInfo(MExpStateInfo_t highlightedStateInfo)
{
    CBRPlanner_StateInfo_t *stateInfo = NULL;

    stateInfo = (CBRPlanner_StateInfo_t *)&highlightedStateInfo;

    pthread_mutex_lock(&socketMutex_);
    sendCommandToServer_(CBRPCOM_CMD_HIGHLIGHTED_STATE);
    sendStateInfo_(*stateInfo);
    pthread_mutex_unlock(&socketMutex_);
}

/**********************************************************************
 * $Log: cfgedit_cbrclient.cc,v $
 * Revision 1.1.1.1  2006/07/12 13:37:54  endo
 * MissionLab 7.0
 *
 * Revision 1.8  2006/02/19 17:52:51  endo
 * Experiment related modifications
 *
 * Revision 1.7  2006/02/14 02:27:18  endo
 * gAutomaticExecution flag and its capability added.
 *
 * Revision 1.6  2006/01/30 02:47:27  endo
 * AO-FNC CBR-CNP Type-I check-in.
 *
 * Revision 1.5  2005/08/12 21:49:58  endo
 * More improvements for August demo.
 *
 * Revision 1.4  2005/08/09 19:12:43  endo
 * Things improved for the August demo.
 *
 * Revision 1.3  2005/07/31 03:39:42  endo
 * Robot ID specified by CNP can be now recognized by mlab during execution.
 *
 * Revision 1.2  2005/07/16 08:47:49  endo
 * CBR-CNP integration
 *
 * Revision 1.1  2005/02/07 22:25:25  endo
 * Mods for usability-2004
 *
 * Revision 1.2  2003/04/06 08:48:14  endo
 * Updated for CBR Wizard Prototype II.
 *
 * Revision 1.1  2002/01/13 01:08:36  endo
 * Initial revision
 *
 **********************************************************************/
